#!/usr/bin/env python
# v1.1

# This script is runs on both the reference and base stations.
# slight mods aug1 2010 for base

# Usage: python gwstation.py [-dt] -[B|R]
# -d	- Daemon mode
# -t 	- testing mode
# -B	- Base station mode (enables probes)
# -R	- Reference station mode (disables probes)

# Import required python libs
import datetime
import sys
import subprocess
import os.path
import os
import signal
import random
import time

random.seed()

BASE='base'
REF='ref'

# Process argv
__testing__ = False
__daemonize__ = False
__mode__ = None
for arg in sys.argv[1:]:
	if arg == '-t':
		__testing__ = True
	elif arg == '-d':
		__daemonize__ = True
	elif arg == '-B':
		__mode__ = BASE
	elif arg == '-R':
		__mode__ = REF
	else:
		print "Useage: %s [-dt] -[B|R]"%(sys.argv[0])
		print "\t-d: Daemonize process"
		print "\t-t: Testing mode - inhibits powering off"
		print "\t-B: Base station mode - enables probes"
		print "\t-R: Reference station mode - disables probes"

if __mode__ is None:
	print "No mode set use flags -B or -R"
	sys.exit(1)

# Import required custom libs
import daemon
import gps
import gprs
import Gumsense
import gwschedule
import gwutil
if __mode__ is BASE:
	import gwprobes

# When this script should run
RUN_HOUR = 11
RUN_MINUTE = 59
RUN_SECOND = 00

# Stay connected for this number * the schedule state:
COMMS_CONNECT_TIME = 0

# Give up if we don't get a GPRS connection after this long:
COMMS_TRY_TIME = datetime.timedelta( minutes = 1 )

# Keep comms up if exists
COMMS_STAY_AWAKE_FILE = "/home/root/comms_stayawake"

# Keep system alive if this file exists
STAY_AWAKE_FILE = "/home/root/stayawake"

# Hold execution of gwstation
WAIT_FILE = "/home/root/wait"

# Server the station communicates to
UOS_SERVER = "env.ecs.soton.ac.uk"
UOS_DIR = "/glacsweb/iceland"

# Special file
SPECIAL_FILE = "/media/cf/%s_special"%__mode__
REMOTE_SPECIAL = "http://%s%s/%s_special"%(UOS_SERVER, UOS_DIR, __mode__)

# Where the station should scp the files to.
SSH_SERVER = "glacsweb@%s"%UOS_SERVER
SSH_TARGET_DIR = "~/www/iceland/incoming/"
SSH_IDENT_FILE = "/home/root/.ssh/id_rsa.db"

if __mode__ is BASE:
	SSH_REVERSE_PORT = 6666
elif __mode__ is REF:
	SSH_REVERSE_PORT = 6667

# Remote power directory
POWER_DIR = "http://%s%s/power"%(UOS_SERVER,UOS_DIR)

# Where to post stations local power state
UPLOAD = "%s/upload.php"%POWER_DIR

# Where to get stations override power state
DOWNLOAD = "%s/download.php"%POWER_DIR

# Base dir -- data, queued and archive all relative to this:
BASE_DIR = "/media/cf"
# Data directory
DATA_DIR = "/data"
# Data queued for scp'ing
QUEUED_DIR = "/queued"
# Data that has been scp'ed
ARCHIVE_DIR = "/archive"

# If daemonized print to this file
LOG_FILE = BASE_DIR + DATA_DIR + "/gw%s"%__mode__

if __mode__ is BASE:
	WIRED_PROBE = 11	#19
	probe_ids = [21,23,25]

# Daemonize the process
if __daemonize__:
	daemon.daemonize(LOG_FILE)
	sys.stdout = gwutil.FlushOut(LOG_FILE)
	sys.stderr = gwutil.FlushOut(LOG_FILE)

# Open gumsense
gum = Gumsense.Gumsense('/home/root/gumsense.conf', jobs=True)

# Turn on the turbo light to indicate this script is running
gum.io.turbo.on()

# Waiting due to wait file
while os.path.exists(WAIT_FILE):
	print "Waiting due to wait file."
	time.sleep(10)

# Check for power loss
last_run_time = gwschedule.get_last_run_time()
now = datetime.datetime.utcnow()

print "Awake at %s, last run time was %s"%(now,last_run_time)

if now < last_run_time:
	print "The current time is set in the past relative to the last run."
	print "Assuming the MSP lost power, resetting to state 0 for safety."
	gum.io.gps.on()
	rc = gps.set_system_time()
	gum.io.gps.off()
	if rc == 1:
		print "Failed to set time from GPS, will try tomorrow."
		print "Powering down"
		gwschedule.set_next_wakeup(gum)
		gwutil.fs_sync()
		gum.io.turbo.off()
		gwutil.poweroff()
		sys.exit(1)

	now = datetime.datetime.utcnow()
	gum.time = now
	gwschedule.force_state(gum, gwschedule.CRITICAL)


now = datetime.datetime.utcnow()
run_time = datetime.datetime(now.year, now.month, now.day, RUN_HOUR, RUN_MINUTE, RUN_SECOND)
if not __testing__:
	# Unless we wake up in 2 minutes, turn off.
	if not now > run_time:
		# Up before run time
		if not ( now + datetime.timedelta(minutes=2) ) > run_time:
			print "Woken too early, going to sleep"
			print "Powering down"
			gwschedule.set_next_wakeup(gum)
		#	gwschedule.set_last_run_time()
			gwutil.fs_sync()
			gum.io.turbo.off()
			gwutil.poweroff()
			sys.exit(1)
		else:
			diff = run_time - now
			if diff.seconds > 0:
				print "Up slightly early, waiting for %i seconds"%diff.seconds
				time.sleep(diff.seconds)
				print "Resuming in %s mode"%__mode__
	else:
		# Up after run time
		# Check to see if we already ran today
		if last_run_time.year == run_time.year and last_run_time.month == run_time.month and last_run_time.day == run_time.day:
			print "Already executed today"
			print "Powering down"
			gwschedule.set_next_wakeup(gum)
			gwutil.fs_sync()
			gum.io.turbo.off()
			gwutil.poweroff()
			sys.exit(1)
		else:
			print "Running late!"


if not __testing__:
	gwschedule.set_next_wakeup(gum)

# Get probe data this will only execute at 11:59:30 or later
if __mode__ is BASE:
	random.shuffle(probe_ids)
	
	print "Powering on wired probe %i"%WIRED_PROBE
	gum.power.probe.on()

	if gwprobes.ping( WIRED_PROBE ):
		print "Ping wired probe %i: success"%WIRED_PROBE
		gwprobes.setclock( WIRED_PROBE )
		gwprobes.senreadstore( WIRED_PROBE )
		
		# Finish is 12:01:00
		finish = run_time + datetime.timedelta(minutes=6, seconds=30)
		while not datetime.datetime.utcnow() > finish:
			pos = 0
			for id in probe_ids:
				if gwprobes.ping( id ):
					print "Ping probe %i: success"%id
					gwprobes.status( id )
					gwprobes.getdata( id )
					gwprobes.sleep( id )
					del probe_ids[pos]

				else:
					print "Ping probe %i: failed"%id

				pos = pos + 1
				
				if datetime.datetime.utcnow() > finish:
					break

		if len(probe_ids) > 0:
			print "Failed to contact probes %s"%probe_ids
		
		gwprobes.getdata( WIRED_PROBE )
	else:
		print "Ping wired probe %i: failed"%WIRED_PROBE

	print "Powering off wired probe %i"%WIRED_PROBE
	gum.power.probe.off()


# Record snow measurement
if __mode__ is BASE:
	snow_power = subprocess.Popen(args="echo 'GPIO out clear' > /proc/gpio/GPIO68", 
					shell = True, 
					stdout=sys.stdout, 
					stderr = sys.stderr)
	snow_power.wait()
	
	gum.jobs.snow.modify(interval=300, 
				next_time=5, 
				sensors=[gum.sensors.snow])
	
	time.sleep(10)
	
	snow_power = subprocess.Popen(args="echo 'GPIO out set' > /proc/gpio/GPIO68", 
					shell = True, 
					stdout = sys.stdout, 
					stderr = sys.stderr)
	snow_power.wait()

	gum.data.fetch(resume=False)

	gum.jobs.clear(gum.jobs.snow.number)

	gum.jobs.resume()
elif __mode__ is REF:
	gum.data.fetch()


print "Determining local power state."
gwschedule.set_local_state(gum)

# Fetch GPS log files
if gwschedule.get_state() > gwschedule.LOW:
	print "Fetching GPS log files."
	gwutil.get_gps_logs(gum)

# Sort out copying files
now = datetime.datetime.utcnow()

# Rename data files to be unique, prepend date.
print "Renaming data files to be unique"
for fname in os.listdir(BASE_DIR + DATA_DIR):
	if os.path.isfile("%s/%s" % (BASE_DIR + DATA_DIR, fname)):
		os.rename("%s/%s" % (BASE_DIR + DATA_DIR, fname),
		"%s/%i-%2.2i-%2.2i-%2.2i.%2.2i-%s" % (BASE_DIR + DATA_DIR, 
			now.year, now.month, now.day, now.hour, 
			now.minute, fname) )

# Create the new tar filename:
tarf = "%s_data_%i-%2.2i-%2.2i-%2.2i.%2.2i.tar.gz" % (__mode__, now.year, 
		now.month, now.day, now.hour, now.minute)

# Tar up stuff
p = subprocess.Popen( args = "tar -czf ./%s/%s ./%s" % (QUEUED_DIR, 
			tarf, DATA_DIR),
			cwd = BASE_DIR, shell = True, stdout = sys.stdout, 
			stderr = sys.stderr )
p.wait()

# Delete the tarred files
for fname in os.listdir( BASE_DIR + DATA_DIR ):
	#if os.path.isfile( fname ):
	if os.path.isfile( "%s/%s" % (BASE_DIR + DATA_DIR, fname) ):
		os.remove( "%s/%s" % (BASE_DIR + DATA_DIR, fname) )

print "Tar created and tarred files deleted."

# Continue only if we have the power
if gwschedule.get_state() > gwschedule.CRITICAL:
	print "Powering on GPRS modem"
	
	# Power on the gprs modem
	gum.io.gprs.on()
	
	power_time = datetime.datetime.utcnow()

	# Give the modem a few seconds to power up
	time.sleep(20)
	
	print "Waiting for GPRS registration"
	while not gprs.registered():
		now = datetime.datetime.utcnow()
		if power_time + COMMS_TRY_TIME < now:
			print "Failed to register GPRS - giving up"
			gum.io.gprs.off()
			break
		else:
			print "Still waiting for GPRS registration"
			time.sleep(5)

	dir, state = gum.io.gprs.status()
	if dir == 'output' and state == 'high':
		print "GSM modem registered with GPRS service"
		time.sleep(10)
		gprs.connect()
		
		start_time = datetime.datetime.utcnow()
		print "Waiting for GPRS PPP interface to come up"
		time.sleep(5)
		while not gprs.up():
			now = datetime.datetime.utcnow()
			if start_time + COMMS_TRY_TIME < now:
				print "Failed to get a GPRS connection within required time - giving up"
				gum.io.gprs.off()
				break
			else:
				print "Still waiting for GPRS PPP interface to come up"
				time.sleep(5)

	if gprs.up():
		print "GPRS interface is up"
		
		# Send our power state home
		print "Sending power state home"
		wget = subprocess.Popen(args="wget -O /dev/null -q %s?id=%s\&state=%i"%(UPLOAD,__mode__,gwschedule.get_state()),
					shell=True,
					stdout = sys.stdout,
					stderr = sys.stderr)
		wget.wait()

		# Copy all the tars across:
		print "Sending all queued files to %s:%s"%(UOS_SERVER,SSH_TARGET_DIR)
	
		for fname in os.listdir( BASE_DIR + QUEUED_DIR ):
			# Copy the data:
			cmd = "scp -q -i %s %s %s:%s" % (SSH_IDENT_FILE,
					fname, SSH_SERVER, SSH_TARGET_DIR)
			print cmd
			p = subprocess.Popen( args = cmd,
						cwd = BASE_DIR + QUEUED_DIR,
						shell = True, 
						stdout = sys.stdout, 
						stderr = sys.stderr )
			r = p.wait()

			# Archive the file if the scp was successful
			if r == 0:
				print "Sent %s"%fname
				cmd = "mv %s %s" % ( fname, 
							BASE_DIR + ARCHIVE_DIR )
				print cmd
				m = subprocess.Popen( args = cmd,
						cwd = BASE_DIR + QUEUED_DIR,
						shell = True, 
						stdout = sys.stdout, 
						stderr = sys.stderr )
				m.wait()
			else:
				print "Sending %s failed"%fname

		# Get override power state
		print "Fetching overide power state"
		wget = subprocess.Popen(args="wget -O %s -q %s?id=%s"%(gwschedule.OVERIDE_FILE, DOWNLOAD,__mode__),
					shell=True,
					stdout = sys.stdout,
					stderr = sys.stderr)
		wget.wait()
		
		# Set the new power state.
		print "Setting new power state"
		gwschedule.set_state(gum)

		# Get special file
		print "Fetching special file"
		wget = subprocess.Popen(args="wget -q -O %s %s"%(SPECIAL_FILE, REMOTE_SPECIAL),
					shell=True,
					stdout = sys.stdout,
					stderr = sys.stderr)
		wget.wait()

		# Execute special file
		if os.path.exists(SPECIAL_FILE):
			print "Executing special file"
			special = subprocess.Popen(args="sh %s"%SPECIAL_FILE,
							shell = True,
							stdout = sys.stdout,
							stderr = sys.stderr)
			special.wait()
		
			print "Moving special file to data dir"
			cmd = "mv %s %s"%(SPECIAL_FILE, BASE_DIR + DATA_DIR)
			print cmd
			m = subprocess.Popen( args = cmd,
						shell = True, 
						stdout = sys.stdout, 
						stderr = sys.stderr )
			m.wait()

		# Establish reverse tunnel.
		print "Establishing reverse tunnel to %s:%i"%(SSH_SERVER, SSH_REVERSE_PORT)
		ssh = subprocess.Popen(args="ssh -i %s %s -R %i:localhost:22"%(SSH_IDENT_FILE, SSH_SERVER, SSH_REVERSE_PORT),
					shell = True,
					stdout = open("/dev/null","w"),
					stderr = sys.stderr)

		# Determine comms window length based on batv state
		comms_window_time = COMMS_CONNECT_TIME * gwschedule.get_state()
		print "Staying awake for %i seconds"%comms_window_time
		time.sleep(comms_window_time)

		# Stay connected whilst comms stay awake file exists.
		while os.path.exists( COMMS_STAY_AWAKE_FILE ):
			print "Staying connected because of comms stay awake file"
			time.sleep(10)

		# Comms window closed, remove reverse tunnel.
		os.kill(ssh.pid, signal.SIGTERM)
		ssh.wait()
		
		
		# Disconnect GPRS and turn off modem.
		print "Disconnecting GPRS."
		gprs.disconnect()
		gum.io.gprs.off()
else:
	print "State %i not high enough for GPRS"%gwschedule.get_state()

# While the stayawake file exists,
# inhibit going into the power off
# procedure when not in testing mode.
while os.path.exists( STAY_AWAKE_FILE):
	print "Staying awake because of stay awake file"
	time.sleep(10)

# If the script is not in testing mode
# then set the next wake up time, the
# last time this script executed, sync 
# the file system, turn off the turbo 
# light and finally power off.
if not __testing__:
	print "Powering down"

	gwschedule.set_next_wakeup(gum)
	
	gwschedule.set_last_run_time()
	
	gwutil.fs_sync()
	
gum.io.turbo.off()

if not __testing__:
	gwutil.poweroff()
